## Copyright (C) 2019 John Donoghue <john.donoghue@ieee.org>
## 
## This program is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <https://www.gnu.org/licenses/>.

## -*- texinfo -*- 
## @deftypefn {} {@var{msg} =} midimsg (0)
## @deftypefnx {} {@var{msg} =} midimsg (@var{type} ....)
## @deftypefnx {} {@var{msg} =} midimsg ("note", @var{channel}, @var{note}, @var{velocity}, @var{duration}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("noteon", @var{channel}, @var{note}, @var{velocity}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("noteoff", @var{channel}, @var{note}, @var{velocity}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("programchange", @var{channel}, @var{prog}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("controlchange", @var{channel}, @var{ccnum}, @var{ccval}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("polykeypressure", @var{channel}, @var{note}, @var{keypressure}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("channelpressure", @var{channel}, @var{chanpressure}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("pitchbend", @var{channel}, @var{pitchchange}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("polyon", @var{channel}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("monoon", @var{channel}, @var{monochannels}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("omnion", @var{channel}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("omnioff", @var{channel}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("allsoundoff", @var{channel}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("allnotesoff", @var{channel}, @var{timestamp})
## @deftypefnx {} {@var{msg} =} midimsg ("resetallcontrollers", @var{channel}, @var{timestamp})
## Create a midimsg object
##
## If the input parameter is 0, create an empty midi message object
## Otherwise the first variable if the type of message to create, followed by the additional parameters for the message.
##
## For each message type, the timestamp value if optional.
## @seealso{midifileread, midisend, midireceive}
## @end deftypefn

classdef midimsg
  properties (SetAccess = private, GetAccess = public)
   data = {};
   timestamp = {};
  endproperties
 
  properties (SetAccess = private, GetAccess = public)
    type; 
  endproperties

  methods
    
    function this = midimsg (typev, varargin)
      this.data = {};
      this.timestamp = {};

      if nargin < 1 || (! ischar (typev) && !isscalar (typev))
        error ("Expected midi type or size");
      endif
      if isscalar (typev)
        if typev == 0
          return; 
        else
          error ("Non zero size not supported yet");
        endif
      endif

      typev = lower (typev);
      switch typev
        case "noteon"
          # channel,note,velocity,timestamp
          if nargin < 4
            error ('noteon expects at least channel,note,velocity')
          endif
          # TODO: change channel to be 1 indexed
          chan = this.check_channel(varargin{1})-1;
          note = varargin{2};
          vel = varargin{3};
          timestamp = 0;
          if nargin > 4
            timestamp = varargin{4};
          endif
          this.data{end+1} = uint8 ([bitor(0x90, chan), note, vel]);
          this.timestamp{end+1} = timestamp;
        case "noteoff"
          # channel,note,velocity,timestamp
          if nargin < 4
            error ('noteoff expects at least channel,note,velocity')
          endif
          chan = this.check_channel(varargin{1})-1;
          note = varargin{2};
          vel = varargin{3};
          timestamp = 0;
          if nargin > 4
            timestamp = varargin{4};
          endif
          this.data{end+1} = uint8 ([bitor(0x80, chan), note, vel]);
          this.timestamp{end+1} = timestamp;
        case "note"
          # channel,note,velocity,duration, timestamp
          if nargin < 5
            error ('note expects at least channel,note,velocity,duration')
          endif
          chan = this.check_channel(varargin{1})-1;
          note = varargin{2};
          vel = varargin{3};
          dur = varargin{4};
          timestamp = 0;
          if nargin > 5
            timestamp = varargin{5};
          endif
          this.data{end+1} = uint8 ([bitor(0x90, chan), note, vel]);
          this.timestamp{end+1} = timestamp;

          timestamp = timestamp + dur;
          this.data{end+1} = uint8 ([bitor(0x90, chan), note, 0]);
          this.timestamp{end+1} = timestamp;
        case "programchange"
          # channel,prog, timestamp
          if nargin < 3
            error ('programchange expects at least channel,note,velocity')
          endif
          chan = this.check_channel(varargin{1})-1;
          prog = varargin{2};
          timestamp = 0;
          if nargin > 3
            timestamp = varargin{3};
          endif
          this.data{end+1} = uint8([bitor(0xc0, chan), prog]);
          this.timestamp{end+1} = timestamp;

        case "controlchange"
          # channel,ccnum, ccval,timestamp
          if nargin < 4
            error ('controlchange expects at least channel,ccnum,ccval')
          endif
          chan = this.check_channel(varargin{1})-1;
          ccnum = varargin{2};
          ccval = varargin{3};
          timestamp = 0;
          if nargin > 4
            timestamp = varargin{4};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), ccnum ccval]);
          this.timestamp{end+1} = timestamp;

        case "polykeypressure"
          # channel, note, pressure, timestamp
          if nargin < 4
            error ('polykeypressure expects at least channel,note,keypressure')
          endif
          chan = this.check_channel(varargin{1})-1;
          note = varargin{2};
          pres = varargin{3};
          timestamp = 0;
          if nargin > 4
            timestamp = varargin{4};
          endif
          this.data{end+1} = uint8([bitor(0xa0, chan), note pres]);
          this.timestamp{end+1} = timestamp;

        case "channelpressure"
          # channel, pressure, timestamp
          if nargin < 3
            error ('channelpressure expects at least channel,keypressure')
          endif
          chan = this.check_channel(varargin{1})-1;
          pres = varargin{2};
          timestamp = 0;
          if nargin > 3
            timestamp = varargin{3};
          endif
          this.data{end+1} = uint8([bitor(0xd0, chan), pres]);
          this.timestamp{end+1} = timestamp;

        case "pitchbend"
          # channel, pitch, timestamp
          if nargin < 3
            error ('channelpressure expects at least channel,pitch')
          endif
          chan = this.check_channel(varargin{1})-1;
          pitch = varargin{2};
          timestamp = 0;
          if nargin > 3
            timestamp = varargin{3};
          endif
          this.data{end+1} = uint8([bitor(0xe0, chan), pitch]);
          this.timestamp{end+1} = timestamp;

        case "polyon"
          # channel, timestamp
          if nargin < 2
            error ('polyon expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 127]);
          this.timestamp{end+1} = timestamp;

        case "monoon"
          # channel,  monochan, timestamp
          if nargin < 3
            error ('monoon expects at least channel and monochannels')
          endif
          chan = this.check_channel(varargin{1})-1;
          mono = varargin{2};
          timestamp = 0;
          if nargin > 3
            timestamp = varargin{3};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 126, mono]);
          this.timestamp{end+1} = timestamp;

        case "omnion"
          # channel, timestamp
          if nargin < 2
            error ('omnion expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 125]);
          this.timestamp{end+1} = timestamp;

        case "omnioff"
          # channel, timestamp
          if nargin < 2
            error ('omnioff expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 124]);
          this.timestamp{end+1} = timestamp;

        case "allsoundoff"
          # channel, timestamp
          if nargin < 2
            error ('allsoundoff expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 120]);
          this.timestamp{end+1} = timestamp;

        case "allnotesoff"
          # channel, timestamp
          if nargin < 2
            error ('allnotesoff expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 123]);
          this.timestamp{end+1} = timestamp;

        case "resetallcontrollers"
          # channel, timestamp
          if nargin < 2
            error ('resetallcontrollers expects at least channel')
          endif
          chan = this.check_channel(varargin{1})-1;
          timestamp = 0;
          if nargin > 2
            timestamp = varargin{2};
          endif
          this.data{end+1} = uint8([bitor(0xb0, chan), 121]);
          this.timestamp{end+1} = timestamp;

        otherwise
          error ("Unknown midi type '%s", typev);
      endswitch
       
    endfunction

    function a = horzcat (a, b)
      if !isa(a, 'midimsg') || !isa(b, 'midimsg')
        error ("Cannot concatenate non midimsg elements");
      endif
      for i=1:length (b.data)
        a.data{end+1} = b.data{i};
        a.timestamp{end+1} = b.timestamp{i};
      endfor
    endfunction

    function e = isempty (this)
      e = isempty(this.data);
    endfunction

    function e = length (this)
      e = length(this.data);
    endfunction

    function val = subsref (p, s)
      if isempty(s)
        error ("midimsg.subsref missing index");
      endif

      switch (s(1).type)
        case "()"
          idx = s(1).subs;
          if (numel (idx) != 1)
            error ("@midimsg/subsref: need exactly one index");
          endif
          val = midimsg.createMessage(p.data{idx{1}}, p.timestamp{idx{1}});
        case "."
          switch tolower(s(1).subs)
          case "timestamp"
            if length(p.timestamp) == 1
              val = p.timestamp{1};
            else
              val = p.timestamp;
            endif
          case "msgbytes"
            if length(p.data) == 1
              val = p.data{1};
            else
              val = p.data;
            endif
          case "nummsgbytes"
            if length(p.data) > 0
              val = length(p.data{1});

              if length(p.data) > 1
                for i = 2:length(p.data)
                  val = [val length(p.data{i})];
                endfor
              endif
            else
              val = 0;
            endif
          case "type"
            if length(p.data) > 0
              data = p.data{1};
              val = p.type_str(data(1), data(2)); 
              if length(p.data) > 1
                for i = 2:length(p.data)
                  data = p.data{i};
                  val = {val p.type_str(data(1), data(2))};
                endfor
              endif
            else
              val = "<none>"
            endif
          case "channel"
            if length(p.data) > 0
              data = p.data{1};
              val = bitand(data(1), 0x0F) + 1;
              if length(p.data) > 1
                for i = 2:length(p.data)
                  data = p.data{i};
                  val = [val (bitand(data(1), 0x0F)+1)];
                endfor
              endif
              val = double(val);
            endif
          otherwise
            error("unimplemented midimsg.subsref property '%s'", s(1).subs);
          endswitch
        otherwise
          error("unimplemented midimsg.subsref type");
      endswitch

      if (numel (s) > 1)
        val = subsref (val, s(2:end));
      endif
    endfunction

    function short_disp (this)
      printf("MIDI Msg\n");
    endfunction

    function display (this)
      printf("MIDI message:\n");
      # 2 options here
      # single array of data
      # or a cell contain multiple arrays
      for i=1:length(this.data)
         data =  this.data{i};
         cmd = data(1);
         chan = bitand(cmd, 0x0F) + 1;
         types = this.type_str(cmd, data(2)); 

         printf (" %-10s Channel: %2d", types, chan);
         if strcmp(types, "NoteOn") || strcmp(types, "NoteOff")
           printf (" Note: %3d Velocity: %3d", data(2), data(3));
         endif
         if strcmp(types, "ProgramChange")
           printf (" Program: %3d", data(2));
         endif
         if strcmp(types, "ControlChange")
           printf (" CCNumber: %3d CCValue: %3d", data(2), data(3));
         endif
         if strcmp(types, "PitchBend")
           v = bitshift(-7, uint16(data(3))) + data(2);
           printf (" PitchChange: %d", v);
         endif
         if strcmp(types, "ChannelPressure")
           printf (" ChannelPressure: %3d", data(2));
         endif
         if strcmp(types, "PolyKeyPressure")
           printf (" Note: %3d KeyPressure: %3d", data(2), data(3));
         endif
         if strcmp(types, "LocalControl")
           printf (" LocalControl: %3d", data(3));
         endif
         if strcmp(types, "MonoOn")
           printf (" MonoChannels: %3d", data(3));
         endif

         printf (" Timestamp: %f", this.timestamp{i});
         printf (" [");
         printf (" 0x%02X", data);
         printf(" ]\n");
      endfor
    endfunction
    
%{
    currently not used as we override subref    
    function v = get.type (this, val)
      if nargin > 1
        cmd = val;
      elseif length(this.data) > 0
        cmd = bitand(this.data{1}(1), 0xF0);
      else
       cmd = -1;   
      endif
      
      v = this.type_str(cmd); 
    endfunction
%}

    function d = uint8 (this)
      d = this.data;
    endfunction

  endmethods

  methods (Access = private)
    function c = check_channel(this, chan)
      if !isscalar (chan) || chan < 1 || chan > 16
        error ("expected channel to be a number between 1..16");
      endif
      c = chan;
    endfunction
    function v = type_str (this, cmd, b1)
      cmdgrp = bitand(cmd, 0xF0);
      switch (cmdgrp)
       case 0x80
         v = "NoteOff";             
       case 0x90
         v = "NoteOn";
       case 0xA0
         v = "PolyKeyPressure";
       case 0xB0
         # depends on next byte for actual msg
         if b1 == 120
           v = "AllSoundOff";
         elseif b1 == 121
           v = "ResetAllControllers";
         elseif b1 == 122
           v = "LocalControl"; # off or on
         elseif b1 == 123
           v = "AllNotesOff";
         elseif b1 == 124
           v = "OmniOff";
         elseif b1 == 125
           v = "OmniOn";
         elseif b1 == 126
           v = "MonoOn";
         elseif b1 == 127
           v = "PolyOn";
         else
           v = "ControlChange";
         endif
       case 0xC0
         v = "ProgramChange";             
       case 0xD0
         v = "ChannelPressure";             
       case 0xE0
         v = "PitchBend";             
       case 0xF0
         if cmd == 0xF0
           v = "SystemExclusive";
         elseif cmd == 0xF1
           v = "MIDITimeCodeQuarterFrame";
         elseif cmd == 0xF2
           v = "Song";
         elseif cmd == 0xF3
           v = "SongSelect";
         elseif cmd == 0xF4
           v = "Reserved";
         elseif cmd == 0xF5
           v = "Reserved";
         elseif cmd == 0xF6
           v = "TuneRequest";
         elseif cmd == 0xF7
           v = "EOX";
         elseif cmd == 0xF8
           v = "TimingClock";
         elseif cmd == 0xF9
           v = "Reserved";
         elseif cmd == 0xFA
           v = "Start";
         elseif cmd == 0xFB
           v = "Continue";
         elseif cmd == 0xFC
           v = "Stop";
         elseif cmd == 0xFD
           v = "Reserved";
         elseif cmd == 0xFE
           v = "ActiveSensing";
         elseif cmd == 0xFF
           v = "Reset";
         endif
         # depend  on other bytes ?
       otherwise
         v = "Data";             
      endswitch
      
    endfunction

  endmethods

  methods (Static=true)
    function msg = createMessage (data, ts)
      if nargin < 1 || !isa (data, 'uint8') || !ismatrix (data)
        error ("Expected matrix of uint8 data")
      endif
      if nargin < 2
        ts = 0;
      endif
      msg = midimsg(0);
      msg.data{end+1} = data;
      msg.timestamp{end+1} = ts;
    endfunction
  endmethods

endclassdef

%!fail midimsg('badtype')

%!fail midimsg

%!test
%! a = midimsg(0);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 0);
%! assert(isempty(a));

%! fail midimsg("note", 0, 60, 127, 2)

%!test
%! a = midimsg("note", 1, 60, 127, 2);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 2);
%! assert(!isempty(a));
%! assert(a.channel, [1 1]);
%! a = midimsg("note", 2, 60, 127, 1.2);
%! t = a.type;
%! assert(length(t) == 2);
%! assert(strcmp(t{1}, "NoteOn"))
%! assert(strcmp(t{2}, "NoteOn"))
%! t = a.timestamp;
%! assert(length(t) == 2);
%! assert(t{1}, 0.0)
%! assert(t{2}, 1.2)
%! assert(a.channel, [2 2]);

%!test
%! a = midimsg("noteon", 1, 60, 20);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "NoteOn");
%! assert(a.channel, 1);
%! assert(a.msgbytes, uint8([0x90 0x3C 0x14]));
%! assert(!isempty(a));
%! a = midimsg("noteon", 2, 60, 20);
%! assert(a.channel, 2);
%! assert(a.msgbytes, uint8([0x91 0x3C 0x14]));

%!test
%! a = midimsg("noteoff", 1, 60, 20);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "NoteOff");
%! assert(a.nummsgbytes, 3);
%! assert(!isempty(a));
%! assert(a.msgbytes, uint8([0x80 0x3C 0x14]));

%!test
%! a = midimsg("programchange", 1, 60);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "ProgramChange");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("controlchange", 1, 60, 65);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "ControlChange");
%! assert(a.nummsgbytes, 3);
%! assert(!isempty(a));

%!test
%! a = midimsg("polykeypressure", 1, 60, 65);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "PolyKeyPressure");
%! assert(a.nummsgbytes, 3);
%! assert(!isempty(a));

%!test
%! a = midimsg("pitchbend", 1, 60);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "PitchBend");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("channelpressure", 1, 60);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "ChannelPressure");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("polyon", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "PolyOn");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("monoon", 1, 0);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "MonoOn");
%! assert(a.nummsgbytes, 3);
%! assert(!isempty(a));

%!test
%! a = midimsg("omnion", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "OmniOn");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("omnioff", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "OmniOff");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("allsoundoff", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "AllSoundOff");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("allnotesoff", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "AllNotesOff");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("resetallcontrollers", 1);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "ResetAllControllers");
%! assert(a.nummsgbytes, 2);
%! assert(!isempty(a));

%!test
%! a = midimsg("noteon", 1, 60, 20);
%! assert(isa(a, "midimsg"));
%! assert(length(a) == 1);
%! assert(a.type, "NoteOn");
%! b = midimsg("noteoff", 2, 60, 20, 5.0);
%! assert(isa(b, "midimsg"));
%! assert(length(b) == 1);
%! assert(b.type, "NoteOff");
%! c = [a b];
%! assert(isa(c, "midimsg"));
%! assert(length(c) == 2);
%! assert(c.nummsgbytes, [3 3]);
%! assert(c.channel, [1 2]);
%! assert(c(1).type, "NoteOn");
%! assert(c(1).channel, 1);
%! assert(c(2).type, "NoteOff");
%! assert(c(2).timestamp, 5.0);
%! assert(c(2).channel, 2);
